/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003-2008  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "types.h"
#include "logging.h"
#include "cart.h"
#include "events.h"
#include "fs.h"
#include "hexs19.h"
#include "joystick.h"
#include "keyboard.h"
#include "m6809.h"
#include "machine.h"
#include "module.h"
#include "sam.h"
#include "snapshot.h"
#include "tape.h"
#include "vdisk.h"
#include "vdrive.h"
#include "xroar.h"

#ifdef TRACE
# include "m6809_trace.h"
#endif

int requested_frameskip;
int frameskip;
int noratelimit = 0;

event_t *xroar_ui_events = NULL;
event_t *xroar_machine_events = NULL;

static const char *snapshot_load = NULL;

static struct {
	const char *ext;
	int filetype;
} filetypes[] = {
	{ "VDK", FILETYPE_VDK },
	{ "JVC", FILETYPE_JVC },
	{ "DSK", FILETYPE_JVC },
	{ "DMK", FILETYPE_DMK },
	{ "BIN", FILETYPE_BIN },
	{ "HEX", FILETYPE_HEX },
	{ "CAS", FILETYPE_CAS },
	{ "WAV", FILETYPE_WAV },
	{ "SN",  FILETYPE_SNA },
	{ "ROM", FILETYPE_ROM },
	{ NULL, FILETYPE_UNKNOWN }
};

static void xroar_helptext(void) {
	puts("  -ui MODULE            specify user-interface module (-ui help for a list)");
	puts("  -vo MODULE            specify video module (-vo help for a list)");
	puts("  -ao MODULE            specify audio module (-ao help for a list)");
	puts("  -fskip FRAMES         specify frameskip (default: 0)");
	puts("  -snap FILENAME        load snapshot after initialising");
	puts("  -h, --help            display this help and exit");
	puts("      --version         output version information and exit");
}

int xroar_init(int argc, char **argv) {
	int i;
	requested_frameskip = 0;

	/* Select a UI module then, possibly using lists specified in that
	 * module, select all other modules */
	ui_module = (UIModule *)module_select_by_arg((Module **)ui_module_list, "-ui", argc, argv);
	/* Select file requester module */
	if (ui_module->filereq_module_list != NULL)
		filereq_module_list = ui_module->filereq_module_list;
	filereq_module = (FileReqModule *)module_select_by_arg((Module **)filereq_module_list, "-filereq", argc, argv);
	/* Select video module */
	if (ui_module->video_module_list != NULL)
		video_module_list = ui_module->video_module_list;
	video_module = (VideoModule *)module_select_by_arg((Module **)video_module_list, "-vo", argc, argv);
	/* Select sound module */
	if (ui_module->sound_module_list != NULL)
		sound_module_list = ui_module->sound_module_list;
	sound_module = (SoundModule *)module_select_by_arg((Module **)sound_module_list, "-ao", argc, argv);
	/* Select keyboard module */
	if (ui_module->keyboard_module_list != NULL)
		keyboard_module_list = ui_module->keyboard_module_list;
	keyboard_module = NULL;
	/* Select joystick module */
	if (ui_module->joystick_module_list != NULL)
		joystick_module_list = ui_module->joystick_module_list;
	joystick_module = NULL;
	/* Look for other relevant command-line options */
	for (i = 1; i < argc; i++) {
		if (!strcmp(argv[i], "-fskip") && i+1<argc) {
			requested_frameskip = strtol(argv[++i], NULL, 0);
			if (requested_frameskip < 0)
				requested_frameskip = 0;
		} else if (!strcmp(argv[i], "-snap")) {
			i++;
			if (i >= argc) break;
			snapshot_load = argv[i];
		} else if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help")
				|| !strcmp(argv[i], "-help")) {
			printf("Usage: xroar [OPTION]...\n\n");
			machine_helptext();
			cart_helptext();
			xroar_helptext();
			module_helptext((Module *)ui_module, (Module **)ui_module_list);
			module_helptext((Module *)filereq_module, (Module **)filereq_module_list);
			module_helptext((Module *)video_module, (Module **)video_module_list);
			module_helptext((Module *)sound_module, (Module **)sound_module_list);
			module_helptext((Module *)keyboard_module, (Module **)keyboard_module_list);
			module_helptext((Module *)joystick_module, (Module **)joystick_module_list);
#ifdef TRACE
			m6809_trace_helptext();
#endif
			exit(0);
		} else if (!strcmp(argv[i], "--version")) {
			printf("XRoar " VERSION "\n");
			puts("Copyright (C) 2008 Ciaran Anscomb");
			puts("This is free software.  You may redistribute copies of it under the terms of");
			puts("the GNU General Public License <http://www.gnu.org/licenses/gpl.html>.");
			puts("There is NO WARRANTY, to the extent permitted by law.");
			exit(0);
		}
	}
	machine_getargs(argc, argv);
	cart_getargs(argc, argv);
#ifdef TRACE
	m6809_trace_getargs(argc, argv);
#endif

	/* Initialise everything */
	current_cycle = 0;
	/* ... modules */
	module_init((Module *)ui_module, argc, argv);
	filereq_module = (FileReqModule *)module_init_from_list((Module **)filereq_module_list, (Module *)filereq_module, argc, argv);
	if (filereq_module == NULL && filereq_module_list != NULL) {
		LOG_WARN("No file requester module initialised.\n");
	}
	video_module = (VideoModule *)module_init_from_list((Module **)video_module_list, (Module *)video_module, argc, argv);
	if (video_module == NULL && video_module_list != NULL) {
		LOG_ERROR("No video module initialised.\n");
		return 1;
	}
	sound_module = (SoundModule *)module_init_from_list((Module **)sound_module_list, (Module *)sound_module, argc, argv);
	if (sound_module == NULL && sound_module_list != NULL) {
		LOG_ERROR("No sound module initialised.\n");
		return 1;
	}
	keyboard_module = (KeyboardModule *)module_init_from_list((Module **)keyboard_module_list, (Module *)keyboard_module, argc, argv);
	if (keyboard_module == NULL && keyboard_module_list != NULL) {
		LOG_WARN("No keyboard module initialised.\n");
	}
	joystick_module = (JoystickModule *)module_init_from_list((Module **)joystick_module_list, (Module *)joystick_module, argc, argv);
	if (joystick_module == NULL && joystick_module_list != NULL) {
		LOG_WARN("No joystick module initialised.\n");
	}
	/* ... subsystems */
	keyboard_init();
	joystick_init();
	machine_init();
	/* Reset everything */
	machine_reset(RESET_HARD);
	if (snapshot_load)
		read_snapshot(snapshot_load);
	return 0;
}

void xroar_shutdown(void) {
	machine_shutdown();
	module_shutdown((Module *)joystick_module);
	module_shutdown((Module *)keyboard_module);
	module_shutdown((Module *)sound_module);
	module_shutdown((Module *)video_module);
	module_shutdown((Module *)filereq_module);
	module_shutdown((Module *)ui_module);
}

void xroar_mainloop(void) {
	while (1) {
		while (EVENT_PENDING(UI_EVENT_LIST))
			DISPATCH_NEXT_EVENT(UI_EVENT_LIST);
		m6809_cycle(UI_EVENT_LIST->at_cycle);
	}
}

int xroar_filetype_by_ext(const char *filename) {
	char *ext;
	int i;
	ext = strrchr(filename, '.');
	if (ext == NULL)
		return FILETYPE_UNKNOWN;
	ext++;
	for (i = 0; filetypes[i].ext; i++) {
		if (!strncasecmp(ext, filetypes[i].ext, strlen(filetypes[i].ext)))
			return filetypes[i].filetype;
	}
	return FILETYPE_UNKNOWN;
}

/* Bits set in 'mode' affect behaviour per-filetype:
 *   XROAR_AUTORUN_CAS: try to autorun any CAS file
 *   XROAR_AUTORUN_CART: enable cartridge FIRQ autostarter
 */
int xroar_load_file(const char *filename, int mode) {
	int filetype;
	if (filename == NULL)
		return 1;
	filetype = xroar_filetype_by_ext(filename);
	switch (filetype) {
		case FILETYPE_VDK:
		case FILETYPE_JVC:
		case FILETYPE_DMK:
			vdrive_eject_disk(0);
			return vdrive_insert_disk(0, vdisk_load(filename));
		case FILETYPE_BIN:
			return coco_bin_read(filename);
		case FILETYPE_HEX:
			return intel_hex_read(filename);
		case FILETYPE_SNA:
			return read_snapshot(filename);
		case FILETYPE_ROM:
			return cart_insert(filename, mode & XROAR_AUTORUN_CART);
		case FILETYPE_CAS:
			if (mode & XROAR_AUTORUN_CAS)
				return tape_autorun(filename);
			else
				return tape_open_reading(filename);
		case FILETYPE_WAV:
		default:
			return tape_open_reading(filename);
	}
}
